/*
 *
 * This file is part of the LXARDOSCOPE package.
 *
 * LXARDOSCOPE is an Arduino based oscilloscope for Linux, using the Xforms library.
 *
 * Copyright (C) 2011 Oskar Leuthold
 * 
 * LXARDOSCOPE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * LXARDOSCOPE is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with LXARDOSCOPE; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 */

/*
 *  This file derived from Counter class in Xforms library
 *  Scaler class
 *
 *  This class implements a counter, where the fine and course steps
 *  are executed entirely in user space. The purpose is to allow for adjusting
 *  step size dynamically as a function of the current counter value. This
 *  feature is useful for wide range scales with relative size steps.  
 */

#include <string.h>
#include <stdlib.h>
#include "forms.h"
#include "lxforms.h"

#define SPEC FL_COUNTER_SPEC

/* given each component a name. parts are numbered 01 4 23 */
enum
{
    NONE, OB0 = 1, OB1 = 2, OB2 = 4, OB3 = 8, OB4 = 16, ALL = 31
};

static int timdel;		/* Delay since last value change */

/* Draws a counter */
static void
draw_scaler(FL_OBJECT * ob)
{
    char str[64];
    int i, btype[5];
    SPEC *sp = (SPEC *) ob->spec;

    /* Compute boxtypes if pushed */
    for (i = 0; i < 5; i++)
	if (ob->pushed && FL_IS_UPBOX(ob->boxtype) &&
	    (sp->mouseobj & (1 << i)))
	    btype[i] = FL_TO_DOWNBOX(ob->boxtype);
	else
	    btype[i] = ob->boxtype;

    if (btype[4] == FL_UP_BOX)
	btype[4] = FL_DOWN_BOX;

    /* Compute sizes. Must not leave any gaps otherwise double buffering will
       not work correctly */
	sp->ww[0] = sp->ww[1] = sp->ww[2] = sp->ww[3] = (int)FL_min(0.18f * ob->w, ob->h);
	sp->ww[4] = ob->w - 4 * sp->ww[0];	/* must calculate this way */
	sp->xx[0] = ob->x;
	sp->xx[1] = sp->xx[0] + sp->ww[0];
	sp->xx[4] = sp->xx[1] + sp->ww[1];
	sp->xx[2] = sp->xx[4] + sp->ww[4];
	sp->xx[3] = sp->xx[2] + sp->ww[2];

//    if (sp->filter)
//	strcpy(str, sp->filter(ob, sp->val, sp->prec));
//    else
    if(sp->prec<10) sprintf(str, "%.*f", sp->prec, sp->val);
	else {
//			this prints  1,000.00  (comma and dot) for the file position scaler in LXOSKOPE
//			did not have the patience to figure out how to get the filter working ....
		sprintf(str, "%.*f", 2, sp->val);
		int len=strlen(str);
		if(len>6){
			for(int n=len;n>=len-6;n--) str[n+1]=str[n];
			str[len-6]=',';
		}
	}
    /* Only draw the parts that need to be drawn  */
    if (sp->draw_type & OB0)
    {
	fl_drw_box(btype[0], sp->xx[0], ob->y, sp->ww[0], ob->h,
		   ob->col1, ob->bw);
	fl_drw_text(FL_ALIGN_CENTER, sp->xx[0], ob->y, sp->ww[0], ob->h,
		    ob->col2, 0, 0, "@#<<");
    }

    if ((sp->draw_type & OB1))
    {
	fl_drw_box(btype[1], sp->xx[1], ob->y, sp->ww[1], ob->h, ob->col1,
		   ob->bw);
	fl_drw_text(FL_ALIGN_CENTER, sp->xx[1], ob->y, sp->ww[1], ob->h,
		    ob->col2, 0, 0, "@#<");
    }

    if ((sp->draw_type & OB4))
    {
	fl_drw_box(btype[4], sp->xx[4], ob->y, sp->ww[4], ob->h,
		   ob->col1, ob->bw);
	fl_drw_text(FL_ALIGN_CENTER, sp->xx[4], ob->y, sp->ww[4], ob->h,
		    ob->lcol, ob->lstyle, ob->lsize, str);
    }

    if ((sp->draw_type & OB2))
    {
	fl_drw_box(btype[2], sp->xx[2], ob->y, sp->ww[2], ob->h,
		   ob->col1, ob->bw);
	fl_drw_text(FL_ALIGN_CENTER, sp->xx[2], ob->y, sp->ww[2], ob->h,
		    ob->col2, 0, 0, "@#>");
    }

    if (sp->draw_type & OB3)
    {
	fl_drw_box(btype[3], sp->xx[3], ob->y, sp->ww[3], ob->h,
		   ob->col1, ob->bw);
	fl_drw_text(FL_ALIGN_CENTER, sp->xx[3], ob->y, sp->ww[3], ob->h,
		    ob->col2, 0, 0, "@#>>");
    }

    if (sp->draw_type == ALL)
	fl_draw_object_label_outside(ob);

    sp->draw_type = ALL;
}

static void
calc_mouse_obj(FL_OBJECT * ob, FL_Coord mx, FL_Coord my)
{
    SPEC *sp = (SPEC *) ob->spec;

    sp->mouseobj = NONE;

    if (my < ob->y || my > ob->y + ob->h || mx < ob->x)
	return;

	if (mx < ob->x + sp->ww[0])
	    sp->mouseobj = OB0;
	else if (mx < sp->ww[1] + sp->xx[1])
	    sp->mouseobj = OB1;
	else if (mx < sp->ww[4] + sp->xx[4])
	    sp->mouseobj = OB4;
	else if (mx < sp->ww[2] + sp->xx[2])
	    sp->mouseobj = OB2;
	else if (mx < sp->ww[3] + sp->xx[3])
	    sp->mouseobj = OB3;
}

/* handles an event on ob */
static int
handle_mouse(FL_OBJECT * ob, int event, FL_Coord mx, FL_Coord my)
{
    SPEC *sp = (SPEC *) ob->spec;
    int changeval = 0;

    if (event == FL_RELEASE)
    {
	sp->mouseobj = NONE;
	fl_redraw_object(ob);
	return 0;
    }
    else if (event == FL_PUSH)
    {
	calc_mouse_obj(ob, mx, my);
	if (sp->mouseobj != NONE)
	    changeval = 1;
	timdel = 1;
    }
    else if (event == FL_MOUSE && sp->mouseobj != NONE)
	changeval = (timdel++ > 10 && (timdel & 1) == 0);

    if (changeval)
    {
	double oval = sp->val;
	if (sp->mouseobj == OB0)
            sp->button=-2;
	if (sp->mouseobj == OB1)
            sp->button=-1;
	if (sp->mouseobj == OB2)
            sp->button=1;
	if (sp->mouseobj == OB3)
            sp->button=2;

	if (sp->val != oval)
	{
	    sp->draw_type = sp->mouseobj | OB4;
	    fl_redraw_object(ob);
	}
    }
    return changeval;
}

/* show which button is active by highlighting the button */

static void
show_focus_obj(FL_OBJECT * ob, FL_Coord mx, FL_Coord my)
{
    SPEC *sp = (SPEC *) ob->spec;
    unsigned int oldobj = sp->mouseobj;

    calc_mouse_obj(ob, mx, my);

    // if same object, do nothing 
    if (sp->mouseobj == oldobj)
	return;

    if (sp->mouseobj && sp->mouseobj != OB4 && sp->mouseobj != oldobj)
    {
	FL_COLOR old = ob->col1;
	sp->draw_type = sp->mouseobj;
	ob->col1 = FL_MCOL;
	fl_redraw_object(ob);
	sp->draw_type = oldobj;
	ob->col1 = old;
	fl_redraw_object(ob);
    }
    else if ((sp->mouseobj == NONE || sp->mouseobj == OB4) && oldobj)
    {
	sp->draw_type = oldobj;
	fl_redraw_object(ob);
    }
}


/* Handles an event */
static int
handle_scaler(FL_OBJECT * ob, int event, FL_Coord mx, FL_Coord my,
	       int key, void *ev)
{
    SPEC *sp = (SPEC *) ob->spec;

    switch (event)
    {
    case FL_DRAW:
	draw_scaler(ob);
	break;
    case FL_DRAWLABEL:
	fl_draw_object_label_outside(ob);
	break;
    case FL_PUSH:
	sp->changed = handle_mouse(ob, event, mx, my);

	if (sp->how_return == FL_RETURN_CHANGED && sp->changed)
	{
	    sp->changed = 0;
	    return 1;
	}
	break;
    case FL_MOUSE:
	if (handle_mouse(ob, event, mx, my))
	    sp->changed = 1;

	if (sp->how_return == FL_RETURN_CHANGED && sp->changed)
	{
	    sp->changed = 0;
	    return 1;
	}
	break;
    case FL_RELEASE:
	if (handle_mouse(ob, event, mx, my))
	    sp->changed = 1;
	else if (sp->how_return == FL_RETURN_END_CHANGED && sp->changed)
	    return 1;
	break;
    case FL_ENTER:
    case FL_MOTION:
    case FL_LEAVE:
	show_focus_obj(ob, mx, my);
	break;
    case FL_FREEMEM:
	fl_free(ob->spec);
	break;
    }
    return 0;
}

/*------------------------------*/

/* creates an object */

FL_OBJECT *
fl_create_scaler(int type, FL_Coord x, FL_Coord y, FL_Coord w, FL_Coord h,
		  const char *label)
{
    FL_OBJECT *ob;
    SPEC *sp;
    fl_add_button_class(FL_SCALER, draw_scaler, 0); 
    ob = fl_create_generic_button(FL_SCALER, type, x, y,w, h, label);
    ob = fl_make_object(FL_SCALER, type, x, y, w, h, label, handle_scaler);
    ob->boxtype = FL_SCALER_BOXTYPE;
    ob->col1 = FL_SCALER_COL1;
    ob->col2 = FL_SCALER_COL2;
    ob->align = FL_SCALER_ALIGN;
    ob->lcol = FL_SCALER_LCOL;

    /* scaler has a different default */
    if (ob->bw == FL_BOUND_WIDTH && ob->bw == 3)
	ob->bw = FL_SCALER_BW;

    ob->spec_size = sizeof(SPEC);
    ob->spec = fl_calloc(1, sizeof(SPEC));
    sp = ((SPEC *) (ob->spec));
    sp->min = -1000000.0;
    sp->max = 1000000.0;
    sp->sstep = 0.1;
    sp->lstep = 1.0;
    sp->val = 0.0;
    sp->prec = 1;
    sp->mouseobj = NONE;
    sp->draw_type = ALL;
    sp->how_return = FL_RETURN_END_CHANGED;
    sp->button = 0;
    return ob;
}

/* Adds an object */
FL_OBJECT *
fl_add_scaler(int type, FL_Coord x, FL_Coord y, FL_Coord w, FL_Coord h,
	       const char *label)
{
    FL_OBJECT *ob;

    ob = fl_create_scaler(type, x, y, w, h, label);
    fl_add_object(fl_current_form, ob);
    return ob;
}


void
fl_set_scaler_value(FL_OBJECT * ob, double val)
{
     SPEC *sp;

    sp = (SPEC *) ob->spec;

    if (sp->val != val)
    {
	sp->val = val;
	sp->draw_type = (ob->visible && ob->form->visible) ? OB4 : ALL;
	fl_redraw_object(ob);
    }
}

void
fl_get_scaler_bounds(FL_OBJECT * ob, double *min, double *max)
{
    *min = ((SPEC *) (ob->spec))->min;
    *max = ((SPEC *) (ob->spec))->max;
}

void
fl_set_scaler_bounds(FL_OBJECT * ob, double min, double max)
{
    SPEC *sp;

    sp = (SPEC *) ob->spec;
    if (sp->min != min || sp->max != max)
    {
	sp->min = min;
	sp->max = max;
	fl_redraw_object(ob);
    }
}


void
fl_set_scaler_step(FL_OBJECT * ob, double s, double l)
{
    SPEC *sp = (SPEC *) ob->spec;

    if (sp->sstep != s || sp->lstep != l)
    {
	sp->sstep = s;
	sp->lstep = l;
	fl_redraw_object(ob);
    }
}

void
fl_get_scaler_step(FL_OBJECT * ob, float *s, float *l)
{
    SPEC *sp = (SPEC *) ob->spec;

    *s = (float)sp->sstep;
    *l = (float)sp->lstep;
}

int
fl_get_scaler_button(FL_OBJECT * ob)
{
    return ((SPEC *) (ob->spec))->button;
}

void
fl_set_scaler_precision(FL_OBJECT * ob, int prec)
{
    if (((SPEC *) (ob->spec))->prec != prec)
    {
	((SPEC *) (ob->spec))->prec = prec;
	fl_redraw_object(ob);
    }
}

double
fl_get_scaler_value(FL_OBJECT * ob)
{
    return ((SPEC *) (ob->spec))->val;
}

/* Sets whether to return value all the time */

void
fl_set_scaler_return(FL_OBJECT * ob, int how)
{
    SPEC *sp = (SPEC *) ob->spec;
    sp->how_return = how;
    if (sp->how_return == FL_RETURN_END)
	sp->how_return = FL_RETURN_END_CHANGED;
    else if (sp->how_return == FL_RETURN_ALWAYS)
	sp->how_return = FL_RETURN_CHANGED;
}


void
fl_set_scaler_filter(FL_OBJECT * ob, FL_VAL_FILTER filter)
{
    ((SPEC *) (ob->spec))->filter = filter;
}
